@@ -1,6 +1,8 @@
-from mock import patch, call
+from mock import patch, call, mock_open
 
+import collections
 import six
+import errno
 from shutil import rmtree
 from tempfile import mkdtemp
 from threading import Timer
@@ -100,15 +102,15 @@ MONMAP_DUMP = b"""{
             {
                 "rank": 0,
                 "name": "ip-172-31-13-119",
-                "addr": "172.31.13.119:6789\/0"},
+                "addr": "172.31.13.119:6789\\/0"},
             {
                 "rank": 1,
                 "name": "ip-172-31-24-50",
-                "addr": "172.31.24.50:6789\/0"},
+                "addr": "172.31.24.50:6789\\/0"},
             {
                 "rank": 2,
                 "name": "ip-172-31-33-107",
-                "addr": "172.31.33.107:6789\/0"}
+                "addr": "172.31.33.107:6789\\/0"}
         ]}}"""
 
 CEPH_CLIENT_RELATION = {
@@ -164,6 +166,25 @@ class TestConfig():
         return self.config.get(key)
 
 
+class CephBasicUtilsTests(TestCase):
+    def setUp(self):
+        super(CephBasicUtilsTests, self).setUp()
+        [self._patch(m) for m in [
+            'check_output',
+        ]]
+
+    def _patch(self, method):
+        _m = patch.object(ceph_utils, method)
+        mock = _m.start()
+        self.addCleanup(_m.stop)
+        setattr(self, method, mock)
+
+    def test_enabled_manager_modules(self):
+        self.check_output.return_value = b'{"enabled_modules": []}'
+        ceph_utils.enabled_manager_modules()
+        self.check_output.assert_called_once_with(['ceph', 'mgr', 'module', 'ls'])
+
+
 class CephUtilsTests(TestCase):
     def setUp(self):
         super(CephUtilsTests, self).setUp()
@@ -176,10 +197,14 @@ class CephUtilsTests(TestCase):
             'relation_ids',
             'relation_set',
             'log',
+            'cmp_pkgrevno',
+            'enabled_manager_modules',
         ]]
         # Ensure the config is setup for mocking properly.
         self.test_config = TestConfig()
         self.config.side_effect = self.test_config.get
+        self.cmp_pkgrevno.return_value = 1
+        self.enabled_manager_modules.return_value = []
 
     def _patch(self, method):
         _m = patch.object(ceph_utils, method)
@@ -187,6 +212,87 @@ class CephUtilsTests(TestCase):
         self.addCleanup(_m.stop)
         setattr(self, method, mock)
 
+    def _get_osd_settings_test_helper(self, settings, expected=None):
+        units = {
+            'client:1': ['ceph-iscsi/1', 'ceph-iscsi/2'],
+            'client:3': ['cinder-ceph/0', 'cinder-ceph/3']}
+        self.relation_ids.return_value = units.keys()
+        self.related_units.side_effect = lambda x: units[x]
+        self.relation_get.side_effect = lambda x, y, z: settings[y]
+        if expected:
+            self.assertEqual(
+                ceph_utils.get_osd_settings('client'),
+                expected)
+        else:
+            ceph_utils.get_osd_settings('client'),
+
+    def test_get_osd_settings_all_unset(self):
+        settings = {
+            'ceph-iscsi/1': None,
+            'ceph-iscsi/2': None,
+            'cinder-ceph/0': None,
+            'cinder-ceph/3': None}
+        self._get_osd_settings_test_helper(settings, {})
+
+    def test_get_osd_settings_one_group_set(self):
+        settings = {
+            'ceph-iscsi/1': '{"osd heartbeat grace": 5}',
+            'ceph-iscsi/2': '{"osd heartbeat grace": 5}',
+            'cinder-ceph/0': '{"osd heartbeat interval": 25}',
+            'cinder-ceph/3': '{"osd heartbeat interval": 25}'}
+        self._get_osd_settings_test_helper(
+            settings,
+            {'osd heartbeat interval': 25,
+             'osd heartbeat grace': 5})
+
+    def test_get_osd_settings_invalid_option(self):
+        settings = {
+            'ceph-iscsi/1': '{"osd foobar": 5}',
+            'ceph-iscsi/2': None,
+            'cinder-ceph/0': None,
+            'cinder-ceph/3': None}
+        self.assertRaises(
+            ceph_utils.OSDSettingNotAllowed,
+            self._get_osd_settings_test_helper,
+            settings)
+
+    def test_get_osd_settings_conflicting_options(self):
+        settings = {
+            'ceph-iscsi/1': '{"osd heartbeat grace": 5}',
+            'ceph-iscsi/2': None,
+            'cinder-ceph/0': '{"osd heartbeat grace": 6}',
+            'cinder-ceph/3': None}
+        self.assertRaises(
+            ceph_utils.OSDSettingConflict,
+            self._get_osd_settings_test_helper,
+            settings)
+
+    @patch.object(ceph_utils, 'get_osd_settings')
+    def test_send_osd_settings(self, _get_osd_settings):
+        self.relation_ids.return_value = ['client:1', 'client:3']
+        _get_osd_settings.return_value = {
+            'osd heartbeat grace': 5,
+            'osd heartbeat interval': 25}
+        ceph_utils.send_osd_settings()
+        expected_calls = [
+            call(
+                relation_id='client:1',
+                relation_settings={
+                    'osd-settings': ('{"osd heartbeat grace": 5, '
+                                     '"osd heartbeat interval": 25}')}),
+            call(
+                relation_id='client:3',
+                relation_settings={
+                    'osd-settings': ('{"osd heartbeat grace": 5, '
+                                     '"osd heartbeat interval": 25}')})]
+        self.relation_set.assert_has_calls(expected_calls, any_order=True)
+
+    @patch.object(ceph_utils, 'get_osd_settings')
+    def test_send_osd_settings_bad_settings(self, _get_osd_settings):
+        _get_osd_settings.side_effect = ceph_utils.OSDSettingConflict()
+        ceph_utils.send_osd_settings()
+        self.assertFalse(self.relation_set.called)
+
     def test_validator_valid(self):
         # 1 is an int
         ceph_utils.validator(value=1,
@@ -244,10 +350,9 @@ class CephUtilsTests(TestCase):
         ])
 
     @patch.object(ceph_utils, 'get_cache_mode')
-    @patch.object(ceph_utils, 'ceph_version')
-    def test_pool_remove_writeback_cache_tier(self, ceph_version, cache_mode):
+    def test_pool_remove_writeback_cache_tier(self, cache_mode):
         cache_mode.return_value = 'writeback'
-        ceph_version.return_value = '10.1.1'
+        self.cmp_pkgrevno.return_value = 1
 
         p = ceph_utils.Pool(name='test', service='admin')
         p.remove_cache_tier(cache_pool='cacher')
@@ -291,20 +396,44 @@ class CephUtilsTests(TestCase):
         pg_num = p.get_pgs(pool_size=3, percent_data=0.1)
         self.assertEquals(2, pg_num)
 
+        # Check device_class is passed to get_osds
+        p.get_pgs(pool_size=3, percent_data=90, device_class='nvme')
+        get_osds.assert_called_with('admin', 'nvme')
+
     @patch.object(ceph_utils, 'get_osds')
     def test_replicated_pool_create_old_ceph(self, get_osds):
+        self.cmp_pkgrevno.return_value = -1
         get_osds.return_value = None
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3)
         p.create()
 
         self.check_call.assert_has_calls([
             call(['ceph', '--id', 'admin', 'osd', 'pool', 'create', 'test', str(200)]),
-            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'test', 'size', str(3)])
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'test', 'size', str(3)]),
         ])
+        self.assertEqual(self.check_call.call_count, 2)
+
+    @patch.object(ceph_utils, 'get_osds')
+    def test_replicated_pool_create_luminous_ceph(self, get_osds):
+        self.cmp_pkgrevno.side_effect = [-1, 1]
+        get_osds.return_value = None
+        p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3)
+        p.create()
+
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd',
+                  'pool', 'create', 'test', str(200)]),
+            call(['ceph', '--id', 'admin', 'osd',
+                  'pool', 'set', 'test', 'size', str(3)]),
+            call(['ceph', '--id', 'admin', 'osd', 'pool',
+                  'application', 'enable', 'test', 'unknown'])
+        ])
+        self.assertEqual(self.check_call.call_count, 3)
 
     @patch.object(ceph_utils, 'get_osds')
     def test_replicated_pool_create_small_osds(self, get_osds):
         get_osds.return_value = range(1, 5)
+        self.cmp_pkgrevno.return_value = -1
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3,
                                       percent_data=10)
         p.create()
@@ -318,6 +447,7 @@ class CephUtilsTests(TestCase):
 
     @patch.object(ceph_utils, 'get_osds')
     def test_replicated_pool_create_medium_osds(self, get_osds):
+        self.cmp_pkgrevno.return_value = -1
         get_osds.return_value = range(1, 9)
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3,
                                       percent_data=50)
@@ -328,11 +458,36 @@ class CephUtilsTests(TestCase):
         self.check_call.assert_has_calls([
             call(['ceph', '--id', 'admin', 'osd', 'pool', 'create', 'test',
                   '128']),
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'test', 'size', '3']),
+        ])
+
+    @patch.object(ceph_utils, 'get_osds')
+    def test_replicated_pool_create_autoscaler(self, get_osds):
+        self.enabled_manager_modules.return_value = ['pg_autoscaler']
+        self.cmp_pkgrevno.return_value = 1
+        get_osds.return_value = range(1, 9)
+        p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3,
+                                      percent_data=50)
+        p.create()
+
+        # Using the PG Calc, for 8 OSDs with a size of 3 and 50% of the data
+        # at 100 PGs/OSD, the number of expected placement groups will be 128
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'create', 'test',
+                  '128']),
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'test', 'size', '3']),
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'test',
+                  'target_size_ratio', '0.5']),
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'application', 'enable', 'test', 'unknown']),
+
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set', 'test',
+                 'pg_autoscale_mode', 'on']),
         ])
 
     @patch.object(ceph_utils, 'get_osds')
     def test_replicated_pool_create_large_osds(self, get_osds):
         get_osds.return_value = range(1, 41)
+        self.cmp_pkgrevno.return_value = -1
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3,
                                       percent_data=100)
         p.create()
@@ -348,6 +503,7 @@ class CephUtilsTests(TestCase):
     @patch.object(ceph_utils, 'get_osds')
     def test_replicated_pool_create_xlarge_osds(self, get_osds):
         get_osds.return_value = range(1, 1001)
+        self.cmp_pkgrevno.return_value = -1
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3,
                                       percent_data=100)
         p.create()
@@ -360,15 +516,19 @@ class CephUtilsTests(TestCase):
                   '32768']),
         ])
 
-    def test_replicated_pool_create_failed(self):
+    @patch.object(ceph_utils, 'get_osds')
+    def test_replicated_pool_create_failed(self, get_osds):
+        get_osds.return_value = range(1, 1001)
         self.check_call.side_effect = CalledProcessError(returncode=1,
                                                          cmd='mock',
                                                          output=None)
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3)
         self.assertRaises(CalledProcessError, p.create)
 
+    @patch.object(ceph_utils, 'get_osds')
     @patch.object(ceph_utils, 'pool_exists')
-    def test_replicated_pool_skips_creation(self, pool_exists):
+    def test_replicated_pool_skips_creation(self, pool_exists, get_osds):
+        get_osds.return_value = range(1, 1001)
         pool_exists.return_value = True
         p = ceph_utils.ReplicatedPool(name='test', service='admin', replicas=3)
         p.create()
@@ -384,6 +544,7 @@ class CephUtilsTests(TestCase):
     @patch.object(ceph_utils, 'get_erasure_profile')
     @patch.object(ceph_utils, 'get_osds')
     def test_erasure_pool_create(self, get_osds, erasure_profile):
+        self.cmp_pkgrevno.return_value = 1
         get_osds.return_value = range(1, 60)
         erasure_profile.return_value = {
             'directory': '/usr/lib/x86_64-linux-gnu/ceph/erasure-code',
@@ -396,7 +557,33 @@ class CephUtilsTests(TestCase):
         p.create()
         self.check_call.assert_has_calls([
             call(['ceph', '--id', 'admin', 'osd', 'pool', 'create', 'test',
-                  '2048', '2048', 'erasure', 'default'])
+                  '2048', '2048', 'erasure', 'default']),
+            call(['ceph', '--id', 'admin', 'osd', 'pool',
+                  'application', 'enable', 'test', 'unknown'])
+        ])
+
+    @patch.object(ceph_utils, 'get_erasure_profile')
+    @patch.object(ceph_utils, 'get_osds')
+    def test_erasure_pool_create_autoscaler(self,
+                                            get_osds,
+                                            erasure_profile):
+        self.enabled_manager_modules.return_value = ['pg_autoscaler']
+        self.cmp_pkgrevno.return_value = 1
+        get_osds.return_value = range(1, 60)
+        erasure_profile.return_value = {
+            'directory': '/usr/lib/x86_64-linux-gnu/ceph/erasure-code',
+            'k': '2',
+            'technique': 'reed_sol_van',
+            'm': '1',
+            'plugin': 'jerasure'}
+        p = ceph_utils.ErasurePool(name='test', service='admin',
+                                   percent_data=100)
+        p.create()
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'create', 'test',
+                  '2048', '2048', 'erasure', 'default']),
+            call(['ceph', '--id', 'admin', 'osd', 'pool',
+                  'application', 'enable', 'test', 'unknown']),
         ])
 
     def test_get_erasure_profile_none(self):
@@ -453,9 +640,23 @@ class CephUtilsTests(TestCase):
 
     def test_set_pool_quota(self):
         self.check_call.return_value = 0
-        ceph_utils.set_pool_quota(service='admin', pool_name='data', max_bytes=1024)
+        ceph_utils.set_pool_quota(service='admin', pool_name='data',
+                                  max_bytes=1024)
         self.check_call.assert_has_calls([
-            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set-quota', 'data', 'max_bytes', '1024'])
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set-quota', 'data',
+                  'max_bytes', '1024'])
+        ])
+        ceph_utils.set_pool_quota(service='admin', pool_name='data',
+                                  max_objects=1024)
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set-quota', 'data',
+                  'max_objects', '1024'])
+        ])
+        ceph_utils.set_pool_quota(service='admin', pool_name='data',
+                                  max_bytes=1024, max_objects=1024)
+        self.check_call.assert_has_calls([
+            call(['ceph', '--id', 'admin', 'osd', 'pool', 'set-quota', 'data',
+                  'max_bytes', '1024', 'max_objects', '1024'])
         ])
 
     def test_remove_pool_quota(self):
@@ -465,11 +666,10 @@ class CephUtilsTests(TestCase):
             call(['ceph', '--id', 'admin', 'osd', 'pool', 'set-quota', 'data', 'max_bytes', '0'])
         ])
 
-    @patch.object(ceph_utils, 'ceph_version')
     @patch.object(ceph_utils, 'erasure_profile_exists')
-    def test_create_erasure_profile(self, existing_profile, mock_version):
+    def test_create_erasure_profile(self, existing_profile):
         existing_profile.return_value = True
-        mock_version.return_value = '10.0.0'
+        self.cmp_pkgrevno.return_value = -1
         ceph_utils.create_erasure_profile(service='admin', profile_name='super-profile', erasure_plugin_name='jerasure',
                                           failure_domain='rack', data_chunks=10, coding_chunks=3)
 
@@ -478,7 +678,7 @@ class CephUtilsTests(TestCase):
                'ruleset-failure-domain=' + 'rack', '--force']
         self.check_call.assert_has_calls([call(cmd)])
 
-        mock_version.return_value = '12.1.0'
+        self.cmp_pkgrevno.return_value = 1
         ceph_utils.create_erasure_profile(service='admin', profile_name='super-profile', erasure_plugin_name='jerasure',
                                           failure_domain='rack', data_chunks=10, coding_chunks=3)
 
@@ -489,6 +689,7 @@ class CephUtilsTests(TestCase):
 
     @patch.object(ceph_utils, 'erasure_profile_exists')
     def test_create_erasure_profile_local(self, existing_profile):
+        self.cmp_pkgrevno.return_value = -1
         existing_profile.return_value = False
         ceph_utils.create_erasure_profile(service='admin', profile_name='super-profile', erasure_plugin_name='local',
                                           failure_domain='rack', data_chunks=10, coding_chunks=3, locality=1)
@@ -500,6 +701,7 @@ class CephUtilsTests(TestCase):
 
     @patch.object(ceph_utils, 'erasure_profile_exists')
     def test_create_erasure_profile_shec(self, existing_profile):
+        self.cmp_pkgrevno.return_value = -1
         existing_profile.return_value = False
         ceph_utils.create_erasure_profile(service='admin', profile_name='super-profile', erasure_plugin_name='shec',
                                           failure_domain='rack', data_chunks=10, coding_chunks=3,
@@ -598,20 +800,41 @@ class CephUtilsTests(TestCase):
         self.assertEqual("writeback", cache_mode)
 
     @patch('os.path.exists')
-    def test_create_keyring(self, _exists):
+    def test_add_key(self, _exists):
         """It creates a new ceph keyring"""
         _exists.return_value = False
-        ceph_utils.create_keyring('cinder', 'cephkey')
+        ceph_utils.add_key('cinder', 'cephkey')
         _cmd = ['ceph-authtool', '/etc/ceph/ceph.client.cinder.keyring',
                 '--create-keyring', '--name=client.cinder',
                 '--add-key=cephkey']
         self.check_call.assert_called_with(_cmd)
 
     @patch('os.path.exists')
-    def test_create_keyring_already_exists(self, _exists):
-        """It creates a new ceph keyring"""
+    def test_add_key_already_exists(self, _exists):
+        """It should insert the key into the existing keyring"""
         _exists.return_value = True
-        ceph_utils.create_keyring('cinder', 'cephkey')
+        try:
+            with patch("__builtin__.open", mock_open(read_data="foo")):
+                ceph_utils.add_key('cinder', 'cephkey')
+        except ImportError:  # Python3
+            with patch("builtins.open", mock_open(read_data="foo")):
+                ceph_utils.add_key('cinder', 'cephkey')
+        self.assertTrue(self.log.called)
+        _cmd = ['ceph-authtool', '/etc/ceph/ceph.client.cinder.keyring',
+                '--create-keyring', '--name=client.cinder',
+                '--add-key=cephkey']
+        self.check_call.assert_called_with(_cmd)
+
+    @patch('os.path.exists')
+    def test_add_key_already_exists_and_key_exists(self, _exists):
+        """Nothing should happen, apart from a log message"""
+        _exists.return_value = True
+        try:
+            with patch("__builtin__.open", mock_open(read_data="cephkey")):
+                ceph_utils.add_key('cinder', 'cephkey')
+        except ImportError:  # Python3
+            with patch("builtins.open", mock_open(read_data="cephkey")):
+                ceph_utils.add_key('cinder', 'cephkey')
         self.assertTrue(self.log.called)
         self.check_call.assert_not_called()
 
@@ -658,23 +881,31 @@ class CephUtilsTests(TestCase):
         _mkdir.assert_called_with('/etc/ceph')
         _install.assert_called_with('ceph-common', fatal=True)
 
-    @patch.object(ceph_utils, 'ceph_version')
-    def test_get_osds(self, version):
-        version.return_value = '0.56.2'
+    def test_get_osds(self):
         self.check_output.return_value = json.dumps([1, 2, 3]).encode('UTF-8')
         self.assertEquals(ceph_utils.get_osds('test'), [1, 2, 3])
 
-    @patch.object(ceph_utils, 'ceph_version')
-    def test_get_osds_argonaut(self, version):
-        version.return_value = '0.48.3'
-        self.assertEquals(ceph_utils.get_osds('test'), None)
-
-    @patch.object(ceph_utils, 'ceph_version')
-    def test_get_osds_none(self, version):
-        version.return_value = '0.56.2'
+    def test_get_osds_none(self):
         self.check_output.return_value = json.dumps(None).encode('UTF-8')
         self.assertEquals(ceph_utils.get_osds('test'), None)
 
+    def test_get_osds_device_class(self):
+        self.check_output.return_value = json.dumps([1, 2, 3]).encode('UTF-8')
+        self.assertEquals(ceph_utils.get_osds('test', 'nvme'), [1, 2, 3])
+        self.check_output.assert_called_once_with(
+            ['ceph', '--id', 'test',
+             'osd', 'crush', 'class',
+             'ls-osd', 'nvme', '--format=json']
+        )
+
+    def test_get_osds_device_class_older(self):
+        self.check_output.return_value = json.dumps([1, 2, 3]).encode('UTF-8')
+        self.cmp_pkgrevno.return_value = -1
+        self.assertEquals(ceph_utils.get_osds('test', 'nvme'), [1, 2, 3])
+        self.check_output.assert_called_once_with(
+            ['ceph', '--id', 'test', 'osd', 'ls', '--format=json']
+        )
+
     @patch.object(ceph_utils, 'get_osds')
     @patch.object(ceph_utils, 'pool_exists')
     def test_create_pool(self, _exists, _get_osds):
@@ -803,7 +1034,7 @@ class CephUtilsTests(TestCase):
         self.assertEquals(ceph_utils.get_ceph_nodes(), [])
 
     def test_configure(self):
-        self._patch('create_keyring')
+        self._patch('add_key')
         self._patch('create_key_file')
         self._patch('get_ceph_nodes')
         self._patch('modprobe')
@@ -820,7 +1051,7 @@ class CephUtilsTests(TestCase):
             _file.write.assert_called_with(_conf)
             _open.assert_called_with('/etc/ceph/ceph.conf', 'w')
         self.modprobe.assert_called_with('rbd')
-        self.create_keyring.assert_called_with('cinder', 'key')
+        self.add_key.assert_called_with('cinder', 'key')
         self.create_key_file.assert_called_with('cinder', 'key')
 
     def test_image_mapped(self):
@@ -991,8 +1222,8 @@ class CephUtilsTests(TestCase):
         e = self.assertRaises(IOError, ceph_utils.make_filesystem, device,
                               timeout=0)
         self.assertEquals(device, e.filename)
-        self.assertEquals(os.errno.ENOENT, e.errno)
-        self.assertEquals(os.strerror(os.errno.ENOENT), e.strerror)
+        self.assertEquals(errno.ENOENT, e.errno)
+        self.assertEquals(os.strerror(errno.ENOENT), e.strerror)
         self.log.assert_called_with(
             'Gave up waiting on block device %s' % device, level='ERROR')
 
@@ -1060,7 +1291,7 @@ class CephUtilsTests(TestCase):
         self.assertEquals(False, ceph_utils.ensure_ceph_keyring(service='foo'))
 
     @patch.object(ceph_utils, '_keyring_path')
-    @patch.object(ceph_utils, 'create_keyring')
+    @patch.object(ceph_utils, 'add_key')
     @patch.object(ceph_utils, 'relation_ids')
     def test_ensure_ceph_keyring_no_relation_but_key(self, rids,
                                                      create, _path):
@@ -1071,7 +1302,7 @@ class CephUtilsTests(TestCase):
         _path.assert_called_with('foo')
 
     @patch.object(ceph_utils, '_keyring_path')
-    @patch.object(ceph_utils, 'create_keyring')
+    @patch.object(ceph_utils, 'add_key')
     @patch.object(ceph_utils, 'relation_ids')
     @patch.object(ceph_utils, 'related_units')
     @patch.object(ceph_utils, 'relation_get')
@@ -1099,26 +1330,6 @@ class CephUtilsTests(TestCase):
             '/etc/ceph/client.foo.keyring'
         ])
 
-    @patch('os.path.exists')
-    def test_ceph_version_not_installed(self, path):
-        path.return_value = False
-        self.assertEquals(ceph_utils.ceph_version(), None)
-
-    @patch.object(ceph_utils, 'check_output')
-    @patch('os.path.exists')
-    def test_ceph_version_error(self, path, output):
-        path.return_value = True
-        output.return_value = b''
-        self.assertEquals(ceph_utils.ceph_version(), None)
-
-    @patch.object(ceph_utils, 'check_output')
-    @patch('os.path.exists')
-    def test_ceph_version_ok(self, path, output):
-        path.return_value = True
-        output.return_value = \
-            b'ceph version 0.67.4 (ad85b8bfafea6232d64cb7ba76a8b6e8252fa0c7)'
-        self.assertEquals(ceph_utils.ceph_version(), '0.67.4')
-
     @patch.object(ceph_utils, 'service_name')
     @patch.object(ceph_utils, 'uuid')
     def test_ceph_broker_rq_class(self, uuid, service_name):
@@ -1126,9 +1337,12 @@ class CephUtilsTests(TestCase):
         uuid.uuid1.return_value = 'uuid'
         rq = ceph_utils.CephBrokerRq()
         rq.add_op_create_pool('pool1', replica_count=1)
+        rq.add_op_create_pool('pool1', replica_count=1)
+        rq.add_op_create_pool('pool2')
         rq.add_op_create_pool('pool2')
         rq.add_op_create_pool('pool3', group='test')
         rq.add_op_request_access_to_group(name='test')
+        rq.add_op_request_access_to_group(name='test')
         rq.add_op_request_access_to_group(name='objects',
                                           key_name='test')
         rq.add_op_request_access_to_group(
@@ -1466,6 +1680,22 @@ class CephUtilsTests(TestCase):
                                     "'unknown': {'blah': 1}}")
         self.assertEqual({'osd': {'foo': 1}}, ctxt)
 
+    @patch.object(ceph_utils, 'get_osd_settings')
+    @patch.object(ceph_utils, 'config')
+    def test_ceph_osd_conf_context_conflict(self, mock_config,
+                                            mock_get_osd_settings):
+        mock_config.return_value = "{'osd': {'osd heartbeat grace': 20}}"
+        mock_get_osd_settings.return_value = {
+            'osd heartbeat grace': 25,
+            'osd heartbeat interval': 5}
+        ctxt = ceph_utils.CephOSDConfContext()()
+        self.assertEqual(ctxt, {
+            'osd': collections.OrderedDict([('osd heartbeat grace', 20)]),
+            'osd_from_client': collections.OrderedDict(
+                [('osd heartbeat interval', 5)]),
+            'osd_from_client_conflict': collections.OrderedDict(
+                [('osd heartbeat grace', 25)])})
+
     @patch.object(ceph_utils, 'local_unit', lambda: "nova-compute/0")
     def test_is_broker_action_done(self):
         tmpdir = mkdtemp()
@@ -1481,7 +1711,7 @@ class CephUtilsTests(TestCase):
                 action = 'restart_nova_compute'
                 ret = ceph_utils.is_broker_action_done(action, rid="ceph:1",
                                                        unit="ceph/0")
-                self.relation_get.assert_has_calls([call('ceph:1', 'ceph/0')])
+                self.relation_get.assert_has_calls([call(rid='ceph:1', unit='ceph/0')])
                 self.assertFalse(ret)
 
                 ceph_utils.mark_broker_action_done(action)
@@ -1493,6 +1723,28 @@ class CephUtilsTests(TestCase):
             if os.path.exists(tmpdir):
                 shutil.rmtree(tmpdir)
 
+    def test_has_broker_rsp(self):
+        rq_id = "3d03e9f6-4c36-11e7-89ba-fa163e7c7ec6"
+        broker_key = ceph_utils.get_broker_rsp_key()
+        self.relation_get.return_value = {broker_key:
+                                          json.dumps({"request-id":
+                                                      rq_id,
+                                                      "exit-code": 0})}
+        ret = ceph_utils.has_broker_rsp(rid="ceph:1", unit="ceph/0")
+        self.assertTrue(ret)
+        self.relation_get.assert_has_calls([call(rid='ceph:1', unit='ceph/0')])
+
+        self.relation_get.return_value = {'something_else':
+                                          json.dumps({"request-id":
+                                                      rq_id,
+                                                      "exit-code": 0})}
+        ret = ceph_utils.has_broker_rsp(rid="ceph:1", unit="ceph/0")
+        self.assertFalse(ret)
+
+        self.relation_get.return_value = None
+        ret = ceph_utils.has_broker_rsp(rid="ceph:1", unit="ceph/0")
+        self.assertFalse(ret)
+
     @patch.object(ceph_utils, 'local_unit', lambda: "nova-compute/0")
     def test_mark_broker_action_done(self):
         tmpdir = mkdtemp()
@@ -1508,9 +1760,103 @@ class CephUtilsTests(TestCase):
                 ceph_utils.mark_broker_action_done(action, rid="ceph:1",
                                                    unit="ceph/0")
                 key = 'unit_0_ceph_broker_action.{}'.format(action)
-                self.relation_get.assert_has_calls([call('ceph:1', 'ceph/0')])
+                self.relation_get.assert_has_calls([call(rid='ceph:1', unit='ceph/0')])
                 kvstore = Storage(db_path)
                 self.assertEqual(kvstore.get(key=key), rq_id)
         finally:
             if os.path.exists(tmpdir):
                 shutil.rmtree(tmpdir)
+
+    def test_add_op_create_replicated_pool(self):
+        base_op = {'app-name': None,
+                   'group': None,
+                   'group-namespace': None,
+                   'max-bytes': None,
+                   'max-objects': None,
+                   'name': 'apool',
+                   'op': 'create-pool',
+                   'pg_num': None,
+                   'replicas': 3,
+                   'weight': None}
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool')
+        self.assertEqual(rq.ops, [base_op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_pool('apool', replica_count=42)
+        op = base_op.copy()
+        op['replicas'] = 42
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', pg_num=42)
+        op = base_op.copy()
+        op['pg_num'] = 42
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', weight=42)
+        op = base_op.copy()
+        op['weight'] = 42
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', group=51)
+        op = base_op.copy()
+        op['group'] = 51
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', namespace='sol-iii')
+        op = base_op.copy()
+        op['group-namespace'] = 'sol-iii'
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', app_name='earth')
+        op = base_op.copy()
+        op['app-name'] = 'earth'
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', max_bytes=42)
+        op = base_op.copy()
+        op['max-bytes'] = 42
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_replicated_pool('apool', max_objects=42)
+        op = base_op.copy()
+        op['max-objects'] = 42
+        self.assertEqual(rq.ops, [op])
+
+    def test_add_op_create_erasure_pool(self):
+        base_op = {'app-name': None,
+                   'erasure-profile': None,
+                   'group': None,
+                   'max-bytes': None,
+                   'max-objects': None,
+                   'name': 'apool',
+                   'op': 'create-pool',
+                   'pool-type': 'erasure',
+                   'weight': None}
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_erasure_pool('apool')
+        self.assertEqual(rq.ops, [base_op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_erasure_pool('apool', weight=42)
+        op = base_op.copy()
+        op['weight'] = 42
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_erasure_pool('apool', group=51)
+        op = base_op.copy()
+        op['group'] = 51
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_erasure_pool('apool', app_name='earth')
+        op = base_op.copy()
+        op['app-name'] = 'earth'
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_erasure_pool('apool', max_bytes=42)
+        op = base_op.copy()
+        op['max-bytes'] = 42
+        self.assertEqual(rq.ops, [op])
+        rq = ceph_utils.CephBrokerRq()
+        rq.add_op_create_erasure_pool('apool', max_objects=42)
+        op = base_op.copy()
+        op['max-objects'] = 42
+        self.assertEqual(rq.ops, [op])
